/************************************************************************************************************************/
/* TECHMAN ROBOT INC. ("Techman") CONFIDENTIAL                                                                          */
/* Copyright (C) 2016~2024 Techman. All Rights Reserved.                                                                */
/* This source code and any compilation or derivative thereof is proprietary and confidential to Techman.               */
/* Under no circumstances may any portion of the source code or any modified version of the source code be distributed, */
/* disclosed or otherwise made available to any third party without the express written consent of Techman.             */
/************************************************************************************************************************/


// The Demo shows that how to build Http server in C++ for TM Vision External Inspection

#include "cpp-httplib\httplib.h"

//The source code of jsoncpp library could be found here: https://github.com/oftc/jsoncpp
#include "json/json.h"

#include <string>
#include <vector>
#include <sstream>

using namespace std;

class CLS_Response_Content
{
public:
  CLS_Response_Content(const string& message, const string& result, const float score) : message(message), result(result), score(score) {};
  std::string message;
  std::string result;
  float score;

  std::string serialize()
  {
    Json::Value out;

    out["message"] = message;
    out["result"] = result;
    out["score"] = score;

    cout << out.toStyledString();
    return out.toStyledString();
  }
};

class DET_Response_Content
{
public:
  class Annotation
  {
  public:
    unsigned int box_cx;
    unsigned int box_cy;
    unsigned int box_h;
    unsigned int box_w;
    std::string label;
    float rotation;
    float score;

    Annotation(const unsigned int box_cx, const unsigned int box_cy, const unsigned int box_h, const unsigned int box_w, const std::string label, const float rotation, const float score)
      : box_cx(box_cx), box_cy(box_cy), box_h(box_h), box_w(box_w), label(label), rotation(rotation), score(score)
    {
    }

    Json::Value serialize()
    {
      Json::Value out;
      out["box_cx"] = box_cx;
      out["box_cy"] = box_cy;
      out["box_h"] = box_h;
      out["box_w"] = box_w;
      out["label"] = label;
      out["rotation"] = rotation;
      out["score"] = score;

      return out;
    }
  };

  std::string message;
  std::string result;
  std::vector<Annotation> annotations;

  std::string serialize()
  {
    Json::Value out;
    out["message"] = message;
    out["result"] = result;

    for (auto annotation : annotations)
    {
      out["annotations"].append(annotation.serialize());
    }

    cout << out;
    return out.toStyledString();
  }
};

/*!
 * Functions/Variables in this name space is for demonstration purpose only,
 */
namespace example
{
  const string classification_api_pattern = "/api/CLS";
  const string detection_api_pattern = "/api/DET";

  CLS_Response_Content CLS_processor(const string model_id, const string image_buf, size_t buffer_size) {

    //Implementation...
    //You may need to convert image_buf to more handy image data type for processing

    return CLS_Response_Content("success", "NG", 0.987);
  };

  DET_Response_Content DET_processor() {
    DET_Response_Content result;

    result.message = "success";
    result.result = "";
    result.annotations.emplace_back(150, 150, 100, 100, "apple", -45, 0.964);
    result.annotations.emplace_back(550, 550, 100, 100, "car", 0, 1.000);
    result.annotations.emplace_back(350, 350, 150, 150, "mobilephone", 135, 0.886);

    return result;
  };
}

int main()
{
  //prepare server
  httplib::Server svr;

  //setting request handler for testing: access "http://localhost:4585/hi" with your browser or TmVision Http Setting page
  svr.Get("/hi", [](const httplib::Request &, httplib::Response &res) {
    res.set_content("Hello World!", "text/plain");
  });

  //setting request handler for testing: access "http://localhost:4585/api" with your browser or TmVision Http Setting page
  svr.Get("/api", [](const httplib::Request &, httplib::Response &res) {
    Json::Value result;
    result["result"] = "api";
    result["message"] = "running";
    res.set_content(result.toStyledString(), "text/plain");
  });

  //setting request handler for testing: access "http://localhost:4585/api/status" with your browser or TmVision Http Setting page
  svr.Get(R"(/api/(\w+))", [](const httplib::Request &req, httplib::Response &res) {
    std::string m_method = req.matches[1];
    if (m_method == "status")
    {
      Json::Value result;
      result["result"] = "status";
      result["message"] = "im ok";
      res.set_content(result.toStyledString(), "text/plain");
    }
    else
    {
      Json::Value error_response;
      error_response["message"] = "fail";
      error_response["result"] = "wrong request";
      res.set_content(error_response.toStyledString(), "text/plain");
    }
  });


  //setting Classification request handler
  svr.Post(example::classification_api_pattern, [](const httplib::Request &rqs, httplib::Response &res) {

    //TODO: Checking your the request parameter
    //e.g. model_id is necessary to example::CLS_processor
    if(!rqs.has_file("model_id"))
    {
      cout << "model_id is not set" << endl;

      //setting response with json string
      Json::Value error_response;
      error_response["message"] = "fail";
      error_response["result"] = "model_id required";
      cout << error_response;

      res.set_content(error_response.toStyledString(), "text/plain");
      return;
    }

    auto model_id = rqs.get_file_value("model_id").content;
    //End of customize request parameter checking

    //retrieve image data: will be formatted in jpg or png
    string image_buf = rqs.get_file_value("image").content;
    /*
    //This is an example of decoding the image_buf string into cv::Mat image format through the opencv2 library.
    std::string bytesImage = image_buf;
    std::vector<char> buffer_v(bytesImage.begin(), bytesImage.end());
    cv::Mat decodedImg = cv::imdecode(buffer_v, cv::IMREAD_UNCHANGED);*/

    //output data for debugging
    ofstream writter;
    writter.open("image.jpg", std::ios_base::binary); //could be jpg or png, it's depends on you
    writter.write(image_buf.data(), image_buf.size());
    writter.close();

    //Invoke the processor with the information provide by request
    CLS_Response_Content result_content = example::CLS_processor(model_id, image_buf, image_buf.size());

    //formatting the response, CLS_Response_Content, to Json string
    res.set_content(result_content.serialize(), "text/plain");
  });

  //setting Detection request handler
  svr.Post(example::detection_api_pattern, [](const httplib::Request &rqs, httplib::Response &res) {

    //TODO: Checking your the request parameter

    //retrieve image data: will be formatted in jpg or png
    string image_buf = rqs.get_file_value("image").content;
    /*
    //This is an example of decoding the image_buf string into cv::Mat image format through the opencv2 library.
    std::string bytesImage = image_buf;
    std::vector<char> buffer_v(bytesImage.begin(), bytesImage.end());
    cv::Mat decodedImg = cv::imdecode(buffer_v, cv::IMREAD_UNCHANGED);*/

    //Invoke the processor with the information provide by request
    DET_Response_Content result_content = example::DET_processor();

    //formatting the response, CLS_Response_Content, to Json string
    res.set_content(result_content.serialize(), "text/plain");
  });

  //start server
  svr.listen("0.0.0.0", 4585);

  return 0;
}

